<?php

namespace UnicaenDbImport\CodeGenerator\Oracle\Helper;

use RuntimeException;
use UnicaenDbImport\CodeGenerator\Helper\HelperException;
use UnicaenDbImport\Domain\Operation;
use UnicaenDbImport\Platforms\OraclePlatform;

/**
 * Génération de code SQL autour des tables.
 *
 * Version Oracle.
 *
 * @author Unicaen
 */
class TableHelper extends \UnicaenDbImport\CodeGenerator\Helper\TableHelper
{
    /**
     * Helper constructor.
     */
    public function __construct()
    {
        $this->platform = new OraclePlatform();
    }

    /**
     * @inheritDoc
     */
    public function generateSQLForIntermediateTableCreation(string $destinationTable, string $sourceCodeColumn, array $columns, string $intermediateTable): string
    {
        // les ; perturbent le driver Oracle donc on met ça dans un bloc BEGIN...END
        return
            'BEGIN ' . PHP_EOL .
            parent::generateSQLForIntermediateTableCreation($destinationTable, $sourceCodeColumn, $columns, $intermediateTable) . PHP_EOL .
            'END;';
    }

    /**
     * @param string $diffSourceTableName
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     */
    public function generateSQLForDiffViewCreation(string $diffSourceTableName, string $destinationTable, string $sourceCodeColumn, array $columns): string
    {
        return $this->generateDiffViewCreationSQLSnippet($diffSourceTableName, $destinationTable, $sourceCodeColumn, $columns) . PHP_EOL;
    }

    /**
     * @param string $diffSourceTableName
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     */
    protected function generateDiffViewCreationSQLSnippet(string $diffSourceTableName, string $destinationTable, string $sourceCodeColumn, array $columns): string
    {
        $name = $this->generateDiffViewName($destinationTable);

        $res = "CREATE OR REPLACE VIEW $name AS " . PHP_EOL;
        $res .= $this->indent(4, $this->generateDiffViewBodySelectSQLSnippet($diffSourceTableName, $destinationTable, $sourceCodeColumn, $columns)) . PHP_EOL;

        return $res;
    }

    /**
     * {@inheritDoc}
     */
    protected function generateDiffViewDeletionSQLSnippet(string $destinationTable): string
    {
        $name = $this->generateDiffViewName($destinationTable);

        return "DROP VIEW $name";
    }

    /**
     * {@inheritDoc}
     */
    public function generateSQLForUpdateOperationInDestinationTable(string $destinationTable, string $sourceCodeColumn, array $columns, string $where = null): string
    {
        return $this->generateSQLUpdate(Operation::OPERATION_UPDATE, $destinationTable, $sourceCodeColumn, $columns, $where);
    }

    /**
     * {@inheritDoc}
     */
    public function generateSQLForUndeleteOperationInDestinationTable(string $destinationTable, string $sourceCodeColumn, array $columns, string $where = null): string
    {
        return $this->generateSQLUpdate(Operation::OPERATION_UNDELETE, $destinationTable, $sourceCodeColumn, $columns, $where);
    }

    /**
     * {@inheritDoc}
     */
    public function generateSQLForDeleteOperationInDestinationTable(string $destinationTable, string $sourceCodeColumn, array $columns, string $where = null): string
    {
        return $this->generateSQLUpdate(Operation::OPERATION_DELETE, $destinationTable, $sourceCodeColumn, $columns, $where);
    }

    /**
     * @param string $operation
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array $columns
     * @param string|null $where
     * @return string
     */
    protected function generateSQLUpdate(string $operation, string $destinationTable, string $sourceCodeColumn, array $columns, string $where = null): string
    {
        $diffViewName = $this->generateDiffViewName($destinationTable);
        try {
            $setters = $this->generateSQLUpdateSetters($operation, $columns);
        } catch (HelperException $e) {
            throw new RuntimeException("Erreur rencontrée lors de la génération SQL", null, $e);
        }
        $where = "diff.operation = '$operation' AND d.$sourceCodeColumn = diff.$sourceCodeColumn";

        // instruction de mise à jour de la table destination
        $markup = "MERGE INTO $destinationTable d USING (SELECT * FROM $diffViewName) diff " . PHP_EOL;
        $markup .= "ON ($where) " . PHP_EOL;
        $markup .= "WHEN MATCHED THEN UPDATE SET " . PHP_EOL;
        $markup .= $setters;

        return $markup;
    }

    /**
     * @param string $sequenceName
     * @return string
     */
    public function generateSQLForSequenceNextVal(string $sequenceName): string
    {
        return $sequenceName . '.nextval';
    }

    /**
     * @inheritDoc
     */
    public function generateSQLForCastToString(string $columnName): string
    {
        return $columnName . "||''";
    }
}